"""Word cloud."""

from dataclasses import dataclass
from typing import Any, Optional

import japanize_matplotlib
import MeCab
import PIL.Image
import wordcloud


@dataclass(frozen=True)
class WordCloudResult:
    """Word cloud result."""

    wordcloud: wordcloud.WordCloud

    def _repr_png_(self) -> Any:
        return self.wordcloud.to_image()._repr_png_()

    def to_file(self, filename: str) -> None:
        """
        Write the word cloud image into the given file.

        Parameters
        ----------
        filename : str
            Filename to be written.

        """
        self.wordcloud.to_file(filename)

    def to_image(self) -> PIL.Image.Image:
        """
        Return the PIL image of the word cloud.

        Returns
        -------
        PIL.Image.Image
            Image of the word cloud.
        """
        return self.wordcloud.to_image()  # type: ignore[no-any-return]


def _get_words(text: str) -> str:
    """Extract words suitable for word cloud."""
    words = []
    tagger = MeCab.Tagger()
    node = tagger.parseToNode(text)
    while node:
        a = node.feature.split(",")
        if a[0] == "名詞" and (a[1] == "普通名詞" or a[1] == "固有名詞"):
            words.append(node.surface)
        node = node.next
    return " ".join(words)


class WordCloud:
    """Word cloud."""

    def __init__(
        self,
        *,
        width: int = 480,
        height: int = 320,
        background_color: str = "white",
    ):
        """
        Construct a word cloud.

        Parameters
        ----------
        width : int, default 480
            Width of the canvas.
        height : int, default 320
            Height of the canvas.
        background_color : str, default "white"
            Background color for the word cloud image.
        """
        self._width = width
        self._height = height
        self._background_color = background_color
        self._words: Optional[str] = None

    def _create(self, seed: Optional[int]) -> wordcloud.WordCloud:
        return wordcloud.WordCloud(
            font_path=japanize_matplotlib.get_font_ttf_path(),
            width=self._width,
            height=self._height,
            background_color=self._background_color,
            random_state=seed,
        )

    def fit(self, text: str) -> None:
        """
        Train the model.

        Parameters
        ----------
        text : str
            Text for training.

        """
        text = _get_words(text)
        wc = self._create(1)
        self._words = wc.process_text(text)

    def generate(
        self,
        seed: Optional[int] = None,
    ) -> WordCloudResult:
        """
        Generate a word cloud.

        Parameters
        ----------
        seed : int, optional
            Random seed.

        Returns
        -------
        WordCloudResult
            Generated word cloud.

        """
        if self._words is None:
            raise ValueError("model is not yet trained")

        wc = self._create(seed)
        result = wc.generate_from_frequencies(self._words)
        return WordCloudResult(result)



"""Text loaders."""
import functools
import io
import re
import urllib.request
import warnings
import zipfile


@functools.cache
def _download(url: str) -> bytes:
    """
    Return the file contents downloaded from the given URL.

    Parameters
    ----------
    url : str
        URL for downloading data.

    Returns
    -------
    bytes
        Downloaded data.

    """
    with urllib.request.urlopen(url) as f:  # noqa: S310
        return f.read()  # type: ignore[no-any-return]


def _remove_annotations(text: str) -> str:
    """
    Remove annotations in the text.

    Parameters
    ----------
    text : str
        Input text.

    Returns
    -------
    str
        Processed text.

    """
    text = re.split(r"\-{5,}", text)[2]  # remove header
    text = re.split("底本：", text)[0]  # remove footer
    text = re.sub("｜", "", text)  # beginning of string with ruby
    text = re.sub("《.+?》", "", text)  # ruby
    text = re.sub("［＃.+?］", "", text)  # comments by staff
    text = text.strip()
    return text


class AozoraBunkoFileError(ValueError):
    """Error raised for unexpected format in files on Aozora Bunko."""


class AozoraBunkoFileWarning(Warning):
    """Warning raised for unexpected format in files on Aozora Bunko."""


def load_aozorabunko(author_id: int, work_id: int, *, raw: bool = False) -> str:
    """
    Return text downloaded from Aozora Bunko (GitHub mirror).

    Parameters
    ----------
    author_id : int
        Author ID.
    work_id : int
        Work ID.
    raw : bool, default False
        Whether it returns a raw text or not.

    Returns
    -------
    str
        Downloaded text.

    Example
    -------
    >>> load_aozorabunko(35, 1567)  # doctest: +SKIP
    ... # author_id: 35, work_id: 1567 => "Run, Melos!"

    """
    # Download the corresponding library card.
    card_file = f"{author_id:0=6}/card{work_id}.html"
    card = _download(
        "https://raw.githubusercontent.com/aozorabunko/aozorabunko/master/cards/"
        f"{card_file}"
    ).decode("utf-8")

    # Search for the ZIP file with ruby text from the library card.
    m = re.search(r"\d+_ruby_\d+\.zip", card)
    if not m:
        # Fallback to the ZIP file without ruby text.
        m = re.search(r"\d+_txt_\d+\.zip", card)
        if not m:
            raise AozoraBunkoFileError(f"ZIP file not detected in {card_file}")
    zipname = m.group(0)

    # Download the ZIP file.
    zipdata = zipfile.ZipFile(
        io.BytesIO(
            _download(
                "https://raw.githubusercontent.com/"
                "aozorabunko/aozorabunko/master/cards/"
                f"{author_id:0=6}/files/{zipname}"
            )
        )
    )

    # Find text files in the ZIP file.
    text_filenames = [s for s in zipdata.namelist() if s.lower().endswith(".txt")]
    if not text_filenames:
        raise AozoraBunkoFileError(f"text file not found in {zipname}")
    if len(text_filenames) >= 2:
        warnings.warn(
            f"more than one text files found: {text_filenames}. Take the first one",
            AozoraBunkoFileWarning,
            stacklevel=2,
        )
    filename = text_filenames[0]

    # Extract text from the ZIP file.
    text = zipdata.read(filename).decode("shift-jis")

    if not raw:
        text = _remove_annotations(text)

    return text



"""Toolbox for Japanese text."""

from ._version import __version__  # noqa: F401
from .loader import load_aozorabunko
from .textmodel import TextModel
from .wordcloud import WordCloud

__all__ = (
    "load_aozorabunko",
    "TextModel",
    "WordCloud",
)



"""Markov chain model."""

import random
import re
import types
from typing import Any, Optional

import igraph as ig
import markovify
import matplotlib.axes
import matplotlib.pyplot as plt
import MeCab
from markovify.chain import BEGIN, END


def _preprocess_text(text: str) -> list[str]:
    """
    Preprocess the text.

    Parameters
    ----------
    text : str
        Input text.

    Returns
    -------
    list[str]
        Sentences.

    """
    tagger = MeCab.Tagger("-Owakati")

    result = []

    for line in text.splitlines():
        line = line.strip()
        line = line.lstrip("　")
        if not line:
            continue

        # Handle brackets.
        if any(c in line for c in "「」（）"):
            while True:
                # "「...」" -> "..."
                if (
                    line[:1] == "「"
                    and line[-1:] == "」"
                    and "「" not in line[1:-1]
                    and "」" not in line[1:-1]
                ):
                    line = line[1:-1].strip()
                    continue
                # "...「...」" -> "......"
                line1 = re.sub("「([^「」]+)」$", r"\g<1>", line)
                if line1 != line:
                    line = line1.strip()
                    continue
                # "（...）" -> "..."
                if (
                    line[:1] == "（"
                    and line[-1:] == "）"
                    and "（" not in line[1:-1]
                    and "）" not in line[1:-1]
                ):
                    line = line[1:-1].strip()
                    continue
                break
            # Remove possibly incomplete sentences,
            # "...「...」...。..." -> "..."
            while True:
                line1 = re.sub("^[^「」]*「[^「」]*」[^「」。！？!?]*[。！？!?]+", "", line)
                if line1 != line:
                    line = line1.strip()
                    continue
                break
            # Check if brackets are still there.
            # It possibly contains incomplete sentences.
            # TODO: case of "「noun」".
            if any(c in line for c in "「」（）"):
                continue

        if not line:
            continue

        # Remove incomplete sentences.
        line = re.sub("[^。！？!?]+$", "", line)

        if not line:
            continue

        line = re.sub("[。！？!?]+", r"\g<0>.", line)
        line = re.sub(r"。+\.", ".", line)
        line = tagger.parse(line).rstrip()
        line = re.sub(r"\s+\.", ".", line)
        result += line.split(".")

    result = [x.strip() for x in result if x]
    result = [x for x in result if x]
    return result


def _custom_word_split(self: "TextModel", sentence: str) -> list[str]:
    tagger = MeCab.Tagger("-Owakati")
    words = tagger.parse(sentence.rstrip("。")).rstrip().split(" ")
    return words  # type: ignore[no-any-return]


def _custom_word_join(self: "TextModel", words: list[str]) -> str:
    sentence = "".join(words)
    if sentence and sentence[-1] not in "。！？!?":
        sentence = sentence + "。"
    return sentence


def _do_monkey_patch(model: markovify.Text) -> None:
    model.word_split = types.MethodType(_custom_word_split, model)
    model.word_join = types.MethodType(_custom_word_join, model)


class TextModel:
    """Text generation model by Markov chain."""

    def __init__(self, state_size: int = 2) -> None:
        """
        Construct a text generation model.

        Parameters
        ----------
        state_size : int, default 2
            Markov chain state size.

        """
        self._state_size = state_size
        self._model: Optional[markovify.Text] = None
        self._compiled_model: Optional[markovify.Text] = None

    def fit(self, text: str) -> None:
        """
        Train the model.

        Parameters
        ----------
        text : str
            Text for training.

        """
        # Build a model.
        model = markovify.Text(
            _preprocess_text(text), state_size=self._state_size, well_formed=False
        )

        self._model = model
        self._compiled_model = None

    def generate(
        self,
        *,
        beginning: Optional[str] = None,
        strict: Optional[bool] = None,
        tries: Optional[int] = None,
        min_words: Optional[int] = None,
        max_words: Optional[int] = None,
        seed: Optional[int] = None,
    ) -> Optional[str]:
        """
        Generate a sentence.

        Parameters
        ----------
        beginning : str, optional
            Beginning word.
        strict : bool, optional
            Whether to strictly use the beginning word or not.
        tries : int, optional
            Number of tries to generate a valid sentence.
        min_words : int, optional
            Minimum number of words.
        max_words : int, optional
            Maximum number of words.
        seed : int, optional
            Random seed.

        Returns
        -------
        str or None
            Generated sentence.

        """
        if self._model is None:
            raise ValueError("model is not yet trained")

        if seed is not None:
            random.seed(seed)

        if self._compiled_model is None:
            self._compiled_model = self._model.compile()
            _do_monkey_patch(self._compiled_model)

        args: dict[str, Any] = {}
        if beginning is not None:
            args["beginning"] = beginning
        if strict is not None:
            args["strict"] = strict
        if tries is not None:
            args["tries"] = tries
        if min_words is not None:
            args["min_words"] = min_words
        if max_words is not None:
            args["max_words"] = max_words

        m = self._compiled_model
        if beginning is None:
            s = m.make_sentence(**args)
        else:
            s = m.make_sentence_with_start(**args)
        return s  # type: ignore[no-any-return]

    def plot(
        self,
        *,
        width: Optional[float] = None,
        height: Optional[float] = None,
        dpi: Optional[int] = None,
        layout: Optional[str] = None,
        seed: Optional[int] = None,
    ) -> matplotlib.axes.Axes:
        """
        Plot the model.

        Parameters
        ----------
        width: float, optional
            Width of the figure.
        height: float, optional
            Height of the figure.
        dpi: int, optional
            Resolution of the figure.
        layout: str, optional
            Layout of the graph.
        seed : int, optional
            Random seed.

        Returns
        -------
        matplotlib.axes.Axes :
            Matplotlib axes containing the plot.

        """
        if self._model is None:
            raise ValueError("model is not yet trained")

        if width is None:
            width = 10
        if height is None:
            height = width

        if seed is not None:
            random.seed(seed)

        m = self._model.chain.model
        n = self._model.state_size
        beginning_state = (BEGIN,) * n
        ending_state = (END,)

        vertices: dict[tuple[str, ...], int] = {}
        edges = []
        weights = []

        def get_state_name(state: tuple[str, ...]) -> str:
            if all(x == BEGIN for x in state):
                return "BEGIN"
            if END in state:
                return "END"
            s = list(state)
            while s[0] == BEGIN:
                del s[0]
            return ",".join(s)

        def traverse(state: tuple[str, ...]) -> None:
            if state in vertices:
                return
            vertices[state] = len(vertices)

            if state not in m:
                return

            total_weight = sum(m[state].values())

            for next_word, weight in m[state].items():
                if next_word == END:
                    next_state = ending_state
                else:
                    next_state = state[1:] + (next_word,)  # type: ignore[assignment]
                traverse(next_state)
                edges.append((vertices[state], vertices[next_state]))
                weights.append(weight / total_weight)

        traverse(beginning_state)

        _, ax = plt.subplots(figsize=(width, height), dpi=dpi)

        g = ig.Graph(len(vertices), edges, directed=True)
        ig.plot(
            g,
            target=ax,
            layout=layout,
            vertex_label=[get_state_name(x) for x in vertices],
            vertex_color="#f99",
            vertex_frame_color="#999",
            edge_label=[f"{x:.02f}" for x in weights],
            edge_color="#999",
            edge_label_color="#33f",
            edge_width=0.5,
            edge_arrow_width=10,
            edge_arrow_size=10,
            edge_align_label=True,
        )

        return ax



"""Versioning."""

import importlib.metadata

try:
    __version__ = importlib.metadata.version("shikaku")
except importlib.metadata.PackageNotFoundError:
    __version__ = None  # type: ignore[assignment]
